"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const fs_1 = __importDefault(require("fs"));
const path_1 = require("path");
const crypto_1 = require("crypto");
const better_sqlite3_1 = __importDefault(require("better-sqlite3"));
const serialize_javascript_1 = __importDefault(require("serialize-javascript"));
const lodash_1 = require("lodash");
const pkgdata = require('../package.json');
class EnmapError extends Error {
    constructor(message, name) {
        super();
        Error.captureStackTrace(this, this.constructor);
        this.name = name || 'EnmapError';
        this.message = message;
    }
}
/*
           ██████╗██╗      █████╗ ███████╗███████╗
          ██╔════╝██║     ██╔══██╗██╔════╝██╔════╝
          ██║     ██║     ███████║███████╗███████╗
          ██║     ██║     ██╔══██║╚════██║╚════██║
          ╚██████╗███████╗██║  ██║███████║███████║
          ╚═════╝╚══════╝╚═╝  ╚═╝╚══════╝╚══════╝
*/
class Enmap {
    // derived from user options
    #verbose;
    #wal;
    #name;
    #serializer;
    #deserializer;
    #autoEnsure;
    #ensureProps;
    // internal references only
    #off = Symbol('option_off');
    #getPaginatedStmt;
    #db;
    #isDestroyed;
    constructor(options = {}) {
        this.#autoEnsure = options.autoEnsure ?? this.#off;
        this.#ensureProps = options.ensureProps ?? true;
        this.#verbose = options.verbose ? options.verbose : () => { };
        this.#wal = options.wal ?? true;
        this.#name = this.#parseName(options.name);
        this.#serializer = options.serializer ? options.serializer : (data) => data;
        this.#deserializer = options.deserializer
            ? options.deserializer
            : (data) => data;
        if (!options.name || options?.name === '::memory::') {
            this.#name = 'MemoryEnmap';
            process.emitWarning('An Enmap was initialised without a name and will be stored in memory only. NO DATA IS SAVED ON DISK.');
            this.#db = new better_sqlite3_1.default(':memory:'); // { verbose: this.#verbose } ?
        }
        else {
            if (!options.dataDir) {
                if (!fs_1.default.existsSync('./data')) {
                    fs_1.default.mkdirSync('./data');
                }
            }
            const dataDir = (0, path_1.resolve)(process.cwd(), options.dataDir || 'data');
            this.#db = new better_sqlite3_1.default(`${dataDir}${path_1.sep}enmap.sqlite`, {
                verbose: this.#verbose,
            });
        }
        if (!this.#db) {
            throw new EnmapError('Database not opened. See error above this one', 'EnmapDBConnectionError');
        }
        const table = this.#db
            .prepare("SELECT count(*) FROM sqlite_master WHERE type='table' AND name = ?;")
            .get(this.#name);
        if (!table['count(*)']) {
            this.#db
                .prepare(`CREATE TABLE ${this.#name} (key text, path text, value text, PRIMARY KEY('key','path'))`)
                .run();
            this.#db.pragma('synchronous = 1');
            if (this.#wal)
                this.#db.pragma('journal_mode = wal');
        }
        this.#db
            .prepare(`CREATE TABLE IF NOT EXISTS 'internal::autonum' (enmap TEXT PRIMARY KEY, lastnum INTEGER)`)
            .run();
        // CLOSE DATABASE ON EXIT TO PREVENT DATA LOSS
        // I wish there was a better way to do this.
        process.on('exit', () => {
            this.close();
        });
        this.#getPaginatedStmt = this.#db.prepare(`SELECT ROWID, * FROM '${this.#name}' WHERE rowid > @lastRowId AND path = '::NULL::' ORDER BY rowid LIMIT @limit;`);
        this.#isDestroyed = false;
    }
    /*
                     ██████╗ ███████╗████████╗
                    ██╔════╝ ██╔════╝╚══██╔══╝
                    ██║  ███╗█████╗     ██║
                    ██║   ██║██╔══╝     ██║
                    ╚██████╔╝███████╗   ██║
                    ╚═════╝ ╚══════╝   ╚═╝
    */
    get name() {
        return this.#name;
    }
    get keys() {
        this.#readyCheck();
        const keys = this.#db.prepare(`SELECT key FROM ${this.#name}`).all();
        return keys.map(res => res.key);
    }
    get values() {
        this.#readyCheck();
        const values = this.#db.prepare(`SELECT value FROM ${this.#name}`).all();
        return values.map(res => this.#parseData(res.key, res.value));
    }
    get entries() {
        this.#readyCheck();
        const entries = this.#db.prepare(`SELECT key, value FROM ${this.#name}`).all();
        return entries.map(res => [this.#parseData(res.value, res.key), res.key]);
    }
    get size() {
        this.#readyCheck();
        const data = this.#db
            .prepare(`SELECT count(*) FROM '${this.#name}';`)
            .get();
        return data['count(*)'];
    }
    get autonum() {
        return (0, crypto_1.randomUUID)();
    }
    get(key, path = '::NULL::') {
        this.#readyCheck();
        if ((0, lodash_1.isNil)(key))
            return null;
        const data = this.#db
            .prepare(`SELECT * FROM ${this.#name} WHERE key = ? AND path = ?;`)
            .get(key, path);
        if (!data)
            return null;
        return this.#parseData(data.key, data.value);
    }
    has(key, path = '::NULL::') {
        this.#readyCheck();
        const data = this.#db.prepare(`SELECT key FROM ${this.#name} WHERE key = ? AND path = ?;`).get(key, path);
        return data?.key === key;
    }
    randomKey(count = 1) {
        this.#readyCheck();
        return this.#db.prepare(`SELECT key FROM  ${this.#name} ORDER BY RANDOM() LIMIT ?;`)
            .all(count)
            .map(res => res.key);
    }
    random(count = 1) {
        this.#readyCheck();
        return this.#db.prepare(`SELECT value, key FROM  ${this.#name} ORDER BY RANDOM() LIMIT ?;`)
            .all(count)
            .map(row => this.#parseData(row.value, row.key));
    }
    /*
                    ███████╗███████╗████████╗
                    ██╔════╝██╔════╝╚══██╔══╝
                    ███████╗█████╗     ██║
                    ╚════██║██╔══╝     ██║
                    ███████║███████╗   ██║
                    ╚══════╝╚══════╝   ╚═╝
                                            
    */
    set(key, value, path) {
        this.#readyCheck();
        if ((0, lodash_1.isNil)(key))
            return this;
        let data = this.get(key);
        if ((0, lodash_1.isNil)(path)) {
            data = value;
        }
        else {
            if ((0, lodash_1.isNil)(data))
                data = {};
            (0, lodash_1.set)(data, path, value);
        }
        const serialized = (0, serialize_javascript_1.default)(this.#serializer(data, key));
        this.#db
            .prepare(`INSERT OR REPLACE INTO ${this.#name} (key, value, path) VALUES (?, ?, '::NULL::');`)
            .run(key, serialized);
        return this;
    }
    update(key, predicate) {
        this.#readyCheck();
        this.#check(key, ['Object']);
        const previousValue = this.get(key);
        const fn = (0, lodash_1.isFunction)(predicate)
            ? predicate
            : () => (0, lodash_1.merge)(previousValue, predicate);
        const merged = fn(previousValue);
        this.set(key, merged);
        return this;
    }
    delete(key, path) {
        this.#readyCheck();
        if ((0, lodash_1.isNil)(key))
            return this;
        this.#db.prepare(`DELETE FROM ${this.#name} WHERE key = ?`)
            .run(key);
        return this;
    }
    clear() {
        this.#readyCheck();
        this.#db.prepare(`DELETE FROM ${this.name}`).run();
        return this;
    }
    math(key, operation, operand, path) {
        this.#readyCheck();
        this.#check(key, 'Number', path);
        const data = this.get(key, path);
        return this.set(key, this.#mathop(data, operation, operand), path);
    }
    inc(key, path) {
        this.#readyCheck();
        this.math(key, '+', 1, path);
    }
    dec(key, path) {
        this.#readyCheck();
        this.math(key, '-', 1, path);
    }
    /*
               █████╗ ██████╗ ██████╗  █████╗ ██╗   ██╗███████╗
              ██╔══██╗██╔══██╗██╔══██╗██╔══██╗╚██╗ ██╔╝██╔════╝
              ███████║██████╔╝██████╔╝███████║ ╚████╔╝ ███████╗
              ██╔══██║██╔══██╗██╔══██╗██╔══██║  ╚██╔╝  ╚════██║
              ██║  ██║██║  ██║██║  ██║██║  ██║   ██║   ███████║
              ╚═╝  ╚═╝╚═╝  ╚═╝╚═╝  ╚═╝╚═╝  ╚═╝   ╚═╝   ╚══════╝
    */
    map(predicate) {
        this.#readyCheck();
        let lastRowId = 0;
        let finished = false;
        const results = [];
        while (!finished) {
            const rows = this.#getPaginatedStmt.all({ lastRowId, limit: 10 });
            rows.forEach(({ key, value }) => {
                const parsedValue = this.#parseData(value, key);
                results.push(predicate(parsedValue, key));
            });
            lastRowId = rows.length ? rows[rows.length - 1].rowid : null;
            if (rows.length < 10)
                finished = true;
        }
        return results;
    }
    filter(predicate) {
        this.#readyCheck();
        let lastRowId = 0;
        let finished = false;
        const results = [];
        while (!finished) {
            const rows = this.#getPaginatedStmt.all({ lastRowId, limit: 10 });
            rows.forEach(({ key, value }) => {
                const parsedValue = this.#parseData(value, key);
                if (predicate(parsedValue, key)) {
                    results.push(parsedValue);
                }
            });
            lastRowId = rows.length ? rows[rows.length - 1].rowid : null;
            if (rows.length < 10)
                finished = true;
        }
        return results;
    }
    find(predicate) {
        this.#readyCheck();
        let lastRowId = 0;
        let finished = false;
        while (!finished) {
            const rows = this.#getPaginatedStmt.all({ lastRowId, limit: 10 });
            const found = rows.find(({ key, value }) => {
                const parsedValue = this.#parseData(value, key);
                return predicate(parsedValue, key);
            });
            if (found) {
                return this.#parseData(found.value, found.key);
            }
            lastRowId = rows.length ? rows[rows.length - 1].rowid : null;
            if (rows.length < 10)
                finished = true;
        }
    }
    findKey(predicate) {
        this.#readyCheck();
        let lastRowId = 0;
        let finished = false;
        while (!finished) {
            const rows = this.#getPaginatedStmt.all({ lastRowId, limit: 10 });
            const found = rows.find(({ key, value }) => {
                const parsedValue = this.#parseData(value, key);
                return predicate(parsedValue, key);
            });
            if (found) {
                return found.key;
            }
            lastRowId = rows.length ? rows[rows.length - 1].rowid : null;
            if (rows.length < 10)
                finished = true;
        }
    }
    some(predicate) {
        this.#readyCheck();
        let lastRowId = 0;
        let finished = false;
        while (!finished) {
            const rows = this.#getPaginatedStmt.all({ lastRowId, limit: 10 });
            const found = rows.some(({ key, value }) => {
                const parsedValue = this.#parseData(value, key);
                return predicate(parsedValue, key);
            });
            if (found) {
                return true;
            }
            lastRowId = rows.length ? rows[rows.length - 1].rowid : null;
            if (rows.length < 10)
                finished = true;
        }
        return false;
    }
    every(predicate) {
        this.#readyCheck();
        let lastRowId = 0;
        let finished = false;
        while (!finished) {
            const rows = this.#getPaginatedStmt.all({ lastRowId, limit: 10 });
            const found = rows.every(({ key, value }) => {
                const parsedValue = this.#parseData(value, key);
                return predicate(parsedValue, key);
            });
            if (!found) {
                return false;
            }
            lastRowId = rows.length ? rows[rows.length - 1].rowid : null;
            if (rows.length < 10)
                finished = true;
        }
        return true;
    }
    reduce(predicate, initialValue) {
        this.#readyCheck();
        let lastRowId = 0;
        let finished = false;
        let accumulator;
        let first = true;
        while (!finished) {
            const rows = this.#getPaginatedStmt.all({ lastRowId, limit: 10 });
            if (first) {
                if (typeof initialValue !== 'undefined') {
                    accumulator = initialValue;
                }
                else {
                    const d = rows.shift();
                    accumulator = this.#parseData(d.value, d.key);
                }
                first = false;
            }
            for (const { key, value } of rows) {
                const parsedValue = this.#parseData(value, key);
                accumulator = predicate(accumulator, parsedValue, key);
            }
            lastRowId = rows.length ? rows[rows.length - 1].rowid : null;
            if (rows.length < 10)
                finished = true;
        }
        return accumulator;
    }
    forEach(predicate) {
        this.#readyCheck();
        const data = this.#db.prepare(`SELECT key, value FROM ${this.#name}`).all();
        data.forEach(({ key, value }) => predicate(this.#parseData(key, value), key));
        let lastRowId = 0;
        let finished = false;
        const results = [];
        while (!finished) {
            const rows = this.#getPaginatedStmt.all({ lastRowId, limit: 10 });
            rows.forEach(({ key, value }) => {
                const parsedValue = this.#parseData(value, key);
                results.push(predicate(parsedValue, key));
            });
            lastRowId = rows.length ? rows[rows.length - 1].rowid : null;
            if (rows.length < 10)
                finished = true;
        }
    }
    push(key, value, path, allowDupes = false) {
        this.#readyCheck();
        const data = this.get(key);
        this.#check(key, 'Array', path);
        if (!(0, lodash_1.isNil)(path)) {
            const propValue = (0, lodash_1.get)(data, path);
            if (!allowDupes && propValue.indexOf(value) > -1)
                return this;
            propValue.push(value);
            (0, lodash_1.set)(data, path, propValue);
        }
        else {
            if (!allowDupes && data.indexOf(value) > -1)
                return this;
            data.push(value);
        }
        return this.set(key, data);
    }
    remove(key, predicate, path) {
        this.#readyCheck();
        this.#check(key, ['Array', 'Object']);
        const data = this.get(key, path);
        const criteria = (0, lodash_1.isFunction)(predicate) ? predicate : (item, index, obj) => item === predicate;
        const index = data.findIndex(criteria);
        if (index > -1) {
            data.splice(index, 1);
        }
        return this.set(key, data, path);
    }
    /*
              ███╗   ███╗██╗███████╗ ██████╗
              ████╗ ████║██║██╔════╝██╔════╝
              ██╔████╔██║██║███████╗██║
              ██║╚██╔╝██║██║╚════██║██║
              ██║ ╚═╝ ██║██║███████║╚██████╗
              ╚═╝     ╚═╝╚═╝╚══════╝ ╚═════╝
    */
    concat(data) {
        this.#readyCheck();
        data.forEach(({ value, key }) => {
            this.set(key, value);
        });
    }
    ensure(key, defaultValue, path = "::NULL::") {
        this.#readyCheck();
        if (this.#autoEnsure !== this.#off) {
            if (!(0, lodash_1.isNil)(defaultValue))
                process.emitWarning(`Saving "${key}" autoEnsure value was provided for this enmap but a default value has also been provided.${' '}The defaultValue will be ignored, autoEnsure value is used instead.`);
            defaultValue = this.#autoEnsure;
        }
        const clonedValue = (0, lodash_1.cloneDeep)(defaultValue);
        if (!(0, lodash_1.isNil)(path)) {
            if (this.#ensureProps)
                this.ensure(key, {});
            if (this.has(key, path))
                return this.get(key, path);
            this.set(key, defaultValue, path);
            return defaultValue;
        }
        if (this.#ensureProps && (0, lodash_1.isObject)(this.get(key))) {
            if (!(0, lodash_1.isObject)(clonedValue))
                throw new EnmapError(`Default value for "${key}" in enmap "${this.#name}" must be an object when merging with an object value.`, 'EnmapArgumentError');
            const merged = (0, lodash_1.merge)(clonedValue, this.get(key));
            this.set(key, merged);
            return merged;
        }
        if (this.has(key))
            return this.get(key);
        this.set(key, clonedValue);
        return clonedValue;
    }
    close() {
        this.#readyCheck();
        this.#db.close();
        return this;
    }
    destroy() {
        this.#readyCheck();
        this.clear();
        this.#isDestroyed = true;
        const transaction = this.#db.transaction((run) => {
            for (const stmt of run) {
                this.#db.prepare(stmt).run();
            }
        });
        transaction([
            `DROP TABLE IF EXISTS ${this.#name};`,
            `DROP TABLE IF EXISTS 'internal::changes::${this.#name}';`,
            `DELETE FROM 'internal::autonum' WHERE enmap = '${this.#name}';`,
        ]);
        return null;
    }
    /**
     * Exports the enmap data to stringified JSON format.
     * **__WARNING__**: Does not work on memory enmaps containing complex data!
     * @returns {string} The enmap data in a stringified JSON format.
     */
    export() {
        this.#readyCheck();
        return (0, serialize_javascript_1.default)({
            name: this.#name,
            version: pkgdata.version,
            exportDate: Date.now(),
            keys: this.map((value, key) => ({ key, value })),
        }, {
            space: 2,
        });
    }
    /**
     * Import an existing json export from enmap from a string. This data must have been exported from enmap,
     * and must be from a version that's equivalent or lower than where you're importing it.
     * @param {string} data The data to import to Enmap. Must contain all the required fields provided by export()
     * @param {boolean} overwrite Defaults to `true`. Whether to overwrite existing key/value data with incoming imported data
     * @param {boolean} clear Defaults to `false`. Whether to clear the enmap of all data before importing
     * (**__WARNING__**: Any existing data will be lost! This cannot be undone.)
     * @returns {Enmap} The enmap with the new data.
     */
    import(data, overwrite = true, clear = false) {
        this.#readyCheck();
        if (clear)
            this.clear();
        if ((0, lodash_1.isNil)(data))
            throw new EnmapError(`No data provided for import() in "${this.#name}"`, 'EnmapImportError');
        try {
            const parsed = eval(`(${data})`);
            for (const thisEntry of parsed.keys) {
                const { key, value } = thisEntry;
                if (!overwrite && this.has(key))
                    continue;
                this.set(key, this.#deserializer(value, key));
            }
        }
        catch (err) {
            console.error(err);
            throw new EnmapError(`Data provided for import() in "${this.#name}" is invalid JSON. Stacktrace:\n${err}`, 'EnmapImportError');
        }
        return this;
    }
    /*
              ██╗███╗   ██╗████████╗███████╗██████╗ ███╗   ██╗ █████╗ ██╗
              ██║████╗  ██║╚══██╔══╝██╔════╝██╔══██╗████╗  ██║██╔══██╗██║
              ██║██╔██╗ ██║   ██║   █████╗  ██████╔╝██╔██╗ ██║███████║██║
              ██║██║╚██╗██║   ██║   ██╔══╝  ██╔══██╗██║╚██╗██║██╔══██║██║
              ██║██║ ╚████║   ██║   ███████╗██║  ██║██║ ╚████║██║  ██║███████╗
              ╚═╝╚═╝  ╚═══╝   ╚═╝   ╚══════╝╚═╝  ╚═╝╚═╝  ╚═══╝╚═╝  ╚═╝╚══════╝
    */
    /*
    * Internal Method. Parses JSON data.
    * Reserved for future use (logical checking)
    * @param {*} data The data to check/parse
    * @returns {*} An object or the original data.
    */
    #parseData(key, value) {
        return this.#deserializer(eval(`(${value})`), key);
    }
    #parseName(name) {
        return name?.toLowerCase().replace(/[\W_]+/g, '-') ?? 'MemoryEnmap';
    }
    /*
    * INTERNAL method to execute a mathematical operation. Cuz... javascript.
    * And I didn't want to import mathjs!
    * @param {number} base the lefthand operand.
    * @param {string} op the operation.
    * @param {number} opand the righthand operand.
    * @return {number} the result.
    */
    #mathop(base, op, opand) {
        // if (base == undefined || op == undefined || opand == undefined)
        //   throw new EnmapError(
        //     'Math Operation requires base and operation',
        //     'EnmapTypeError',
        //   );
        switch (op) {
            case 'add':
            case 'addition':
            case '+':
                return base + opand;
            case 'sub':
            case 'subtract':
            case '-':
                return base - opand;
            case 'mult':
            case 'multiply':
            case '*':
                return base * opand;
            case 'div':
            case 'divide':
            case '/':
                return base / opand;
            case 'exp':
            case 'exponent':
            case '^':
                return Math.pow(base, opand);
            case 'mod':
            case 'modulo':
            case '%':
                return base % opand;
            case 'rand':
            case 'random':
                return Math.floor(Math.random() * Math.floor(opand));
        }
        return null;
    }
    /*
    * INTERNAL method to verify the type of a key or property
    * Will THROW AN ERROR on wrong type, to simplify code.
    * @param {string} key Required. The key of the element to check
    * @param {string} type Required. The javascript constructor to check
    * @param {string} path Optional. The dotProp path to the property in the object enmap.
    */
    #check(key, type, path) {
        if (!this.has(key))
            throw new EnmapError(`The key "${key}" does not exist in the enmap "${this.#name}"`, 'EnmapPathError');
        if (!type)
            return;
        if (!(0, lodash_1.isArray)(type))
            type = [type];
        if (!(0, lodash_1.isNil)(path)) {
            this.#check(key, 'Object');
            const data = this.get(key);
            if ((0, lodash_1.isNil)((0, lodash_1.get)(data, path))) {
                throw new EnmapError(`The property "${path}" in key "${key}" does not exist. Please set() it or ensure() it."`, 'EnmapPathError');
            }
            if (type.includes((0, lodash_1.get)(data, path).constructor.name)) {
                throw new EnmapError(`The property "${path}" in key "${key}" is not of type "${type}.join(
            '" or "',
          )}" in the enmap "${this.#name}" 
(key was of type "${(0, lodash_1.get)(data, path).constructor.name}")`, 'EnmapTypeError');
            }
        }
        else if (!type.includes(this.get(key).constructor.name)) {
            throw new EnmapError(`The value for key "${key}" is not of type "${type}.join(
          '" or "',
        )}" in the enmap "${this.#name}" (value was of type "${this.get(key).constructor.name}")`, 'EnmapTypeError');
        }
    }
    /*
    * Internal Method. Verifies that the database is ready, assuming persistence is used.
    */
    #readyCheck() {
        if (this.#isDestroyed)
            throw new EnmapError('This enmap has been destroyed and can no longer be used without being re-initialized.', 'EnmapDestroyedError');
    }
}
exports.default = Enmap;
